@@ -14,6 +14,7 @@ from compiler.consts import SC_LOCAL, SC_GLOBAL_IMPLICIT, SC_GLOBAL_EXPLICIT, \
 from compiler.consts import (CO_VARARGS, CO_VARKEYWORDS, CO_NEWLOCALS,
      CO_NESTED, CO_GENERATOR, CO_FUTURE_DIVISION,
      CO_FUTURE_ABSIMPORT, CO_FUTURE_WITH_STATEMENT, CO_FUTURE_PRINT_FUNCTION)
+from .visitor import ASTVisitor
 
 from . import config
 
@@ -321,6 +322,11 @@ class CodeGenerator:
             else:
                 self.emit(prefix + '_GLOBAL', name)
         elif scope == SC_FREE or scope == SC_CELL:
+            if isinstance(self.scope, symbols.ClassScope):
+                if prefix == "STORE" and name not in self.scope.nonlocals:
+                    self.emit(prefix + '_NAME', name)
+                    return
+
             if isinstance(self.scope, symbols.ClassScope) and prefix == "LOAD":
                 self.emit(prefix + '_CLASSDEREF', name)
             else:
@@ -406,10 +412,23 @@ class CodeGenerator:
         #    self.setDocstring(node.doc)
         self.storeName(node.name)
 
+    visitAsyncFunctionDef = visitFunctionDef
+
     def visitLambda(self, node):
         self.update_lineno(node)
         self._visitFuncOrLambda(node, isLambda=1)
 
+    def processBody(self, body, gen):
+        walker = ASTVisitor(gen)
+
+        if isinstance(body, list):
+            for stmt in body:
+                if isinstance(stmt, ast.Return):
+                    gen.seen_toplevel_return = True
+                gen.visit(stmt)
+        else:
+            gen.visit(body)
+
     def _visitFuncOrLambda(self, node, isLambda=0):
         if not isLambda and node.decorator_list:
             for decorator in node.decorator_list:
@@ -423,7 +442,9 @@ class CodeGenerator:
         body = node.body
         if not isLambda:
             body = self.skip_docstring(body)
-        walk(body, gen)
+
+        self.processBody(body, gen)
+
         gen.finish()
         for default in node.args.defaults:
             self.visit(default)
@@ -589,6 +610,72 @@ class CodeGenerator:
             self.visit(node.orelse)
         self.nextBlock(after)
 
+    def visitAsyncFor(self, node):
+        start = self.newBlock()
+        anchor = self.newBlock()
+        or_else = self.newBlock()
+        jump_after = self.newBlock()
+        after = self.newBlock()
+
+        self.set_lineno(node)
+
+        self.setups.push((LOOP, start))
+        self.emit('SETUP_LOOP', jump_after)
+        self.visit(node.iter)
+        self.emit('GET_AITER')
+        self.emit('LOAD_CONST', None)
+        self.emit('YIELD_FROM')
+
+        try_body = self.newBlock()
+        handlers = self.newBlock()
+        loop_body = self.newBlock()
+
+        self.nextBlock(start)
+        self.emit('SETUP_EXCEPT', handlers)
+        self.nextBlock(try_body)
+        self.setups.push((EXCEPT, try_body))
+        self.emit('GET_ANEXT')
+        self.emit('LOAD_CONST', None)
+        self.emit('YIELD_FROM')
+        self.visit(node.target)
+        self.emit('POP_BLOCK')
+        self.setups.pop()
+        self.emit('JUMP_FORWARD', loop_body)
+
+        self.startBlock(handlers)
+        self.emit('DUP_TOP')
+        self.emit('LOAD_GLOBAL', 'StopAsyncIteration')
+
+        self.emit('COMPARE_OP', 'exception match')
+        next = self.newBlock()
+        self.emit('POP_JUMP_IF_FALSE', next)
+        self.nextBlock()
+        self.emit('POP_TOP')
+        self.emit('POP_TOP')
+        self.emit('POP_TOP')
+        self.emit('POP_EXCEPT')
+        self.emit('POP_BLOCK')
+        self.emit('JUMP_ABSOLUTE', or_else)
+
+        self.startBlock(next)
+        self.emit('END_FINALLY')
+
+        self.startBlock(loop_body)
+        self.visit(node.body)
+        self.emit('JUMP_ABSOLUTE', start)
+
+        self.nextBlock(anchor)
+        self.emit('POP_BLOCK')
+        self.setups.pop()
+
+        self.nextBlock(jump_after)
+        self.emit('JUMP_ABSOLUTE', after)
+
+        self.startBlock(or_else)
+        if node.orelse:
+            self.visit(node.orelse)
+        self.nextBlock(after)
+
     def visitBreak(self, node):
         if not self.setups:
             raise SyntaxError("'break' outside loop (%s, %d)" % \
@@ -1060,6 +1147,49 @@ class CodeGenerator:
             self.setups.pop()
             self.__with_count -= 1
 
+    def visitAsyncWith(self, node):
+        self.set_lineno(node)
+        body = self.newBlock()
+        stack = []
+        for withitem in node.items:
+            final = self.newBlock()
+            stack.append(final)
+            self.__with_count += 1
+            valuevar = "_[%d]" % self.__with_count
+            self.visit(withitem.context_expr)
+
+            self.emit('BEFORE_ASYNC_WITH')
+            self.emit('GET_AWAITABLE')
+            self.emit('LOAD_CONST', None)
+            self.emit('YIELD_FROM')
+            self.emit('SETUP_ASYNC_WITH', final)
+
+            if withitem.optional_vars is None:
+                self.emit('POP_TOP')
+            else:
+                self.visit(withitem.optional_vars)
+
+            self.setups.push((TRY_FINALLY, body))
+
+        self.nextBlock(body)
+        self.visit(node.body)
+
+        while stack:
+            final = stack.pop()
+            self.emit('POP_BLOCK')
+            self.setups.pop()
+            self.emit('LOAD_CONST', None)
+            self.nextBlock(final)
+            self.setups.push((END_FINALLY, final))
+            self.emit('WITH_CLEANUP_START')
+            self.emit('GET_AWAITABLE')
+            self.emit('LOAD_CONST', None)
+            self.emit('YIELD_FROM')
+            self.emit('WITH_CLEANUP_FINISH')
+            self.emit('END_FINALLY')
+            self.setups.pop()
+            self.__with_count -= 1
+
     # misc
 
     def visitDiscard(self, node):
@@ -1293,19 +1423,26 @@ class CodeGenerator:
         kw = 0
         self.update_lineno(node)
         self.visit(node.func)
-        star_args = None
-        dstar_args = None
+        ex_starargs = 0
+        have_star = False
         for i in range(len(node.args)):
             arg = node.args[i]
             if isinstance(arg, ast.Starred):
-                # TODO: Implement arbitrary arg unpacking, as introduced in Py3.5
-                assert i == len(node.args) - 1
-                star_args = arg
-                self.visit(star_args)
+                have_star = True
+                if ex_starargs != 0 or i != len(node.args) - 1:
+                    ex_starargs += 1
+                self.visit(arg)
             else:
                 self.visit(arg)
-                pos = pos + 1
+                if ex_starargs != 0:
+                    ex_starargs += 1
+                    self.emit('BUILD_TUPLE', 1)
+                else:
+                    pos = pos + 1
+        if ex_starargs != 0:
+            self.emit('BUILD_LIST_UNPACK', ex_starargs)
 
+        dstar_args = None
         for kwarg in node.keywords:
             if kwarg.arg is None:
                 dstar_args = kwarg.value
@@ -1315,7 +1452,6 @@ class CodeGenerator:
                 self.visit(kwarg.value)
                 kw = kw + 1
 
-        have_star = star_args is not None
         have_dstar = dstar_args is not None
         opcode = callfunc_opcode_info[have_star, have_dstar]
         self.emit(opcode, kw << 8 | pos)
@@ -1366,6 +1502,13 @@ class CodeGenerator:
         self.emit('LOAD_CONST', None)
         self.emit('YIELD_FROM')
 
+    def visitAwait(self, node):
+        self.update_lineno(node)
+        self.visit(node.value)
+        self.emit('GET_AWAITABLE')
+        self.emit('LOAD_CONST', None)
+        self.emit('YIELD_FROM')
+
     # slice and subscript stuff
 
     def visitSubscript(self, node, aug_flag=None):
@@ -1458,23 +1601,52 @@ class CodeGenerator:
         else:
             self.emit('UNPACK_SEQUENCE', before)
 
-    def visitTuple(self, node):
+    def hasStarred(self, elts):
+        for elt in elts:
+            if isinstance(elt, ast.Starred):
+                return True
+        return False
+
+    def _visitListOrTuple(self, node, build_op, build_ex_op):
         self.update_lineno(node)
         if isinstance(node.ctx, ast.Store):
             self._visitUnpack(node)
+            starred_load = False
+        else:
+            starred_load = self.hasStarred(node.elts)
+
+        chunks = 0
+        in_chunk = 0
+
+        def out_chunk():
+            nonlocal chunks, in_chunk
+            if in_chunk:
+                self.emit('BUILD_TUPLE', in_chunk)
+                in_chunk = 0
+                chunks += 1
+
         for elt in node.elts:
+            if starred_load:
+                if isinstance(elt, ast.Starred):
+                    out_chunk()
+                    chunks += 1
+                else:
+                    in_chunk += 1
             self.visit(elt)
+        # Output trailing chunk, if any
+        out_chunk()
+
         if isinstance(node.ctx, ast.Load):
-            self.emit('BUILD_TUPLE', len(node.elts))
+            if starred_load:
+                self.emit(build_ex_op, chunks)
+            else:
+                self.emit(build_op, len(node.elts))
+
+    def visitTuple(self, node):
+        self._visitListOrTuple(node, 'BUILD_TUPLE', 'BUILD_TUPLE_UNPACK')
 
     def visitList(self, node):
-        self.update_lineno(node)
-        if isinstance(node.ctx, ast.Store):
-            self._visitUnpack(node)
-        for elt in node.elts:
-            self.visit(elt)
-        if isinstance(node.ctx, ast.Load):
-            self.emit('BUILD_LIST', len(node.elts))
+        self._visitListOrTuple(node, 'BUILD_LIST', 'BUILD_LIST_UNPACK')
 
     def visitSet(self, node):
         self.update_lineno(node)
@@ -1516,10 +1688,35 @@ class CodeGenerator:
 
     def visitDict(self, node):
         self.update_lineno(node)
-        for k, v in zip(node.keys, node.values):
-            self.visit(k)
-            self.visit(v)
-        self.emit('BUILD_MAP', len(node.keys))
+        has_dstar = None in node.keys
+        if not has_dstar:
+            for k, v in zip(node.keys, node.values):
+                self.visit(k)
+                self.visit(v)
+            self.emit('BUILD_MAP', len(node.keys))
+        else:
+            chunks = 0
+            in_chunk = 0
+
+            def out_chunk():
+                nonlocal chunks, in_chunk
+                if in_chunk:
+                    self.emit('BUILD_MAP', in_chunk)
+                    chunks += 1
+                    in_chunk = 0
+
+            for k, v in zip(node.keys, node.values):
+                if k is None:
+                    out_chunk()
+                    self.visit(v)
+                    chunks += 1
+                else:
+                    self.visit(k)
+                    self.visit(v)
+                    in_chunk += 1
+
+            out_chunk()
+            self.emit('BUILD_MAP_UNPACK', chunks)
 
 class NestedScopeMixin:
     """Defines initClass() for nested scoping (Python 2.2-compatible)"""
@@ -1594,6 +1791,7 @@ class AbstractFunctionCode:
             name = func.name
 
         self.name = name
+        self.seen_toplevel_return = False
 
         args = [self.mangle(elt.arg) for elt in func.args.args]
         kwonlyargs = [self.mangle(elt.arg) for elt in func.args.kwonlyargs]
@@ -1630,6 +1828,8 @@ class AbstractFunctionCode:
         return self.module
 
     def finish(self):
+        if self.seen_toplevel_return:
+            return
         self.graph.startExitBlock()
         if not self.isLambda:
             self.emit('LOAD_CONST', None)
@@ -1702,7 +1902,8 @@ class AbstractClassCode:
 
     def finish(self):
         self.graph.startExitBlock()
-        if self.scope.check_name('__class__') == SC_CELL:
+        #if self.scope.check_name('__class__') == SC_CELL:
+        if '__class__' in self.scope.cells:
             self.emit('LOAD_CLOSURE', '__class__')
         else:
             self.emit('LOAD_CONST', None)
